#include <windows.h>

//*****************************************************************************
// (c) 1995, Paul Andre LeBlanc. All rights reserved.
//*****************************************************************************
// Visual Bell sound driver.
//
// This project creates a sound driver supplement which enhances MessageBeep
// for aurally impared users. In addition to performing it's normal action,
// when this driver is installed MessageBeep will flash the caption and border
// of the active window.
//
// This driver also broadcasts a registered window message to all top-level
// windows system-wide when a beep is sounded, allowing an external program to
// supplement the single caption flash with additional visual cues. Such a
// program can handle this message in it's main window procedure as follows:
//
// #define WM_BEEP RegisterWindowMessage("VisBellBeep")
//
// LRESULT CALLBACK _export WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
// {
//   switch(msg)
//   {
//     // handle various messages...
//     default:
//	 if (msg == WM_BEEP)
//	 {
//         // this is the window which was flashed
//         HWND flashed = (HWND) wParam;
//         // This is the task which called MessageBeep 
//         HTASK task = (HTASK) LOWORD(lParam);
//         // do something here...
//       }
//   }
//   return DefWindowProc(hwnd, msg, wParam, lParam);
// }   
//
//
// To install this driver on your system you must do the following:
//
// 1) Copy visbell.drv to your windows\system directory.
//
// 2) Open windows\system.ini and look for the line which reads
//    "sound.drv=driver_name", where driver_name is the name of the
//    current sound driver. This line is found in the [boot] section
//    of the file. Change it to read "visbell.sound.drv=driver_name".
//
//    Example:
//    Change this:               to this:
//    sound.drv=mmsound.drv      visbell.sound.drv=mmsound.drv              
//
// 3) Add the following line to system.ini, either just above or just
//    below the line you just modified:
//
//    sound.drv=visbell.drv
//
//*****************************************************************************
// Legal nonsense
//*****************************************************************************
// VISBELL.DRV as distributed in this package is FREEWARE. You are free to
// distribute it to anyone you wish, provided that you do not charge for it.
//
// VISBELL.C is included for illustration only. You may modify it for use in
// your own programs, but you may not use it as-is to recreate VISBELL.DRV
// unless you are fully prepared to accept that such a re-creation will also
// be FREEWARE. If you alter this source to produce an enhanced version of
// VISBELL.DRV which is targetted at providing essentially the same service to
// the hearing-impaired community I ask that you also give serious
// consideration to distributing such an enhanced product as FREEWARE.  
//
// Paul Andre LeBlanc
// Compuserve ID: [71043,1643]

extern HINSTANCE _hInstance;
HMODULE driver = 0;

//*****************************************************************************
// Test if OK to chain to original driver
//*****************************************************************************

#define	OkToCall(x) (x != NULL)

//*****************************************************************************
// DoBeep
//*****************************************************************************
// This is the function eventually called by MessageBeep. The uAlert argument
// is the same as that passed to MessageBeep. 

HWND WINAPI (*_GetActiveWindow)(void) = NULL;
BOOL WINAPI (*_FlashWindow)(HWND, BOOL) = NULL;
BOOL WINAPI (*_PostMessage)(HWND, UINT, WPARAM, LPARAM) = NULL;
UINT WINAPI (*_RegisterWindowMessage)(LPCSTR) = NULL;
void WINAPI (*_DoBeep)(UINT) = NULL;

#define	WM_BEEP	_RegisterWindowMessage("VisBellBeep")

#pragma argsused
void WINAPI _export DoBeep(UINT uAlert)
{
  HWND hwnd;
  if (!OkToCall(_GetActiveWindow))
  {
    // Note: We cannot link directly to USER because of the load order
    // when Windows is starting up. So we have to link indirectly to get
    // access to functions in USER.
    HMODULE user = GetModuleHandle("USER");
    (FARPROC) _GetActiveWindow = GetProcAddress(user, "GETACTIVEWINDOW");
    (FARPROC) _FlashWindow = GetProcAddress(user, "FLASHWINDOW");
    (FARPROC) _PostMessage = GetProcAddress(user, "POSTMESSAGE");
    (FARPROC) _RegisterWindowMessage = GetProcAddress(user, "REGISTERWINDOWMESSAGE");
  }
  hwnd = _GetActiveWindow();
  _FlashWindow(hwnd, TRUE);
  _FlashWindow(hwnd, FALSE);
  _PostMessage(HWND_BROADCAST, WM_BEEP, (WPARAM) hwnd, MAKELPARAM(GetCurrentTask(), 0L));
  if (OkToCall(DoBeep)) _DoBeep(uAlert);
}

//*****************************************************************************
// WaitSoundState
//*****************************************************************************

int WINAPI (*_WaitSoundState)(int) = NULL;

int WINAPI _export WaitSoundState(int fnState)
{
  return (OkToCall(_WaitSoundState))
    ? _WaitSoundState(fnState)
    : 0;
}

//*****************************************************************************
// SetVoiceAccent
//*****************************************************************************

int WINAPI (*_SetVoiceAccent)(int, int, int, int, int) = NULL;

int WINAPI _export SetVoiceAccent(int nVoice, int nTempo, int nVolume, int fnMode, int nPitch)
{
  return (OkToCall(_SetVoiceAccent))
    ? _SetVoiceAccent(nVoice, nTempo, nVolume, fnMode, nPitch)
    : 0;
}

//*****************************************************************************
// SetVoiceQueueSize
//*****************************************************************************

int WINAPI (*_SetVoiceQueueSize)(int, int) = NULL;

int WINAPI _export SetVoiceQueueSize(int nVoice, int cbQueueSize)
{
  return (OkToCall(_SetVoiceQueueSize))
    ? _SetVoiceQueueSize(nVoice, cbQueueSize)
    : 0;
}

//*****************************************************************************
// SetVoiceSound
//*****************************************************************************

int WINAPI (*_SetVoiceSound)(int, DWORD, int) = NULL;

int WINAPI _export SetVoiceSound(int nVoice, DWORD dwFrequency, int nDuration)
{
  return (OkToCall(_SetVoiceSound))
    ? _SetVoiceSound(nVoice, dwFrequency, nDuration)
    : 0;
}

//*****************************************************************************
// StopSound
//*****************************************************************************

int WINAPI (*_StopSound)(void) = NULL;

int WINAPI _export StopSound(void)
{
  return (OkToCall(_StopSound))
    ? _StopSound()
    : 0;
}

//*****************************************************************************
// GetThresholdEvent
//*****************************************************************************

int FAR *WINAPI (*_GetThresholdEvent)(void) = NULL;

int FAR *WINAPI _export GetThresholdEvent(void)
{
  return (OkToCall(_GetThresholdEvent))
    ? _GetThresholdEvent()
    : 0;
}

//*****************************************************************************
// SetVoiceNote
//*****************************************************************************

int WINAPI (*_SetVoiceNote)(int, int, int, int) = NULL;

int WINAPI _export SetVoiceNote(int nVoice, int nValue, int nLength, int cDots)
{
  return (OkToCall(_SetVoiceNote))
    ? _SetVoiceNote(nVoice, nValue, nLength, cDots)
    : 0;
}

//*****************************************************************************
// SetSoundNoise
//*****************************************************************************

int WINAPI (*_SetSoundNoise)(int, int) = NULL;

int WINAPI _export SetSoundNoise(int fnSource, int nDuration)
{
  return (OkToCall(_SetSoundNoise))
    ? _SetSoundNoise(fnSource, nDuration)
    : 0;
}

//*****************************************************************************
// StartSound
//*****************************************************************************

int WINAPI (*_StartSound)(void) = NULL;

int WINAPI _export StartSound(void)
{
  return (OkToCall(_StartSound))
    ? _StartSound()
    : 0;
}

//*****************************************************************************
// SetVoiceThreshold
//*****************************************************************************

int WINAPI (*_SetVoiceThreshold)(int, int) = NULL;

int WINAPI _export SetVoiceThreshold(int nVoice, int cNotesThreshold)
{
  return (OkToCall(_SetVoiceThreshold))
    ? _SetVoiceThreshold(nVoice, cNotesThreshold)
    : 0;
}

//*****************************************************************************
// CloseSound
//*****************************************************************************

void WINAPI (*_CloseSound)(void) = NULL;

void WINAPI _export CloseSound(void)
{
  if (OkToCall(_CloseSound)) CloseSound();
}

//*****************************************************************************
// CountVoiceNotes
//*****************************************************************************

int WINAPI (*_CountVoiceNotes)(int) = NULL;

int WINAPI _export CountVoiceNotes(int nVoice)
{
  return (OkToCall(_CountVoiceNotes))
    ? _CountVoiceNotes(nVoice)
    : 0;
}

//*****************************************************************************
// SetVoiceEnvelope
//*****************************************************************************

int WINAPI (*_SetVoiceEnvelope)(int, int, int) = NULL;

int WINAPI _export SetVoiceEnvelope(int nVoice, int nShape, int nRepeat)
{
  return (OkToCall(_SetVoiceEnvelope))
    ? _SetVoiceEnvelope(nVoice, nShape, nRepeat)
    : 0;
}

//*****************************************************************************
// OpenSound
//*****************************************************************************

int WINAPI (*_OpenSound)(void) = NULL;

int WINAPI _export OpenSound(void)
{
  return (OkToCall(_OpenSound))
    ? _OpenSound()
    : 0;
}

//*****************************************************************************
// SyncAllVoices
//*****************************************************************************

int WINAPI (*_SyncAllVoices)(void) = NULL;

int WINAPI _export SyncAllVoices(void)
{
  return (OkToCall(_SyncAllVoices))
    ? _SyncAllVoices()
    : 0;
}

//*****************************************************************************
// GetThresholdStatus
//*****************************************************************************

int WINAPI (*_GetThresholdStatus)(void) = NULL;

int WINAPI _export GetThresholdStatus(void)
{
  return (OkToCall(_GetThresholdStatus))
    ? _GetThresholdStatus()
    : 0;
}

//*****************************************************************************
// Fiddle with the module name so the drivers won't clash
//*****************************************************************************
// Windows does not allow two loaded modules to have the same module name, but
// in order to work this DLL must have "KEYBOARD" as it's module name - the
// same name as the original keyboard driver. To work around this we modify the
// task database entry for each module at runtime, first renaming this one to
// allow the original keyboard driver to be loaded, then renaming the original
// driver to "_OLDKEYB" and restoring this module to "KEYBOARD".

void SetModuleName(HINSTANCE hInstance, LPSTR name)
{
  HMODULE hModule = SELECTOROF(GlobalLock(GetModuleHandle(MAKELP(0, hInstance))));
  LPSTR ModuleName = MAKELP(hModule, *(WORD FAR *) MAKELP(hModule, 0x0026));
  *ModuleName++ = lstrlen(name);
  hmemcpy(ModuleName, name, lstrlen(name));
}

//*****************************************************************************
// Install the linkage to the original driver
//*****************************************************************************

UINT Install(void)
{
  UINT result = 0;
  static char szDriverName[256];
  GetPrivateProfileString("boot", "visbell.sound.drv", "sound.drv", szDriverName, sizeof(szDriverName), "system.ini");
  SetModuleName(_hInstance, "_VISBELL");
  if (HINSTANCE_ERROR < (driver = LoadLibrary(szDriverName)))
  {
    (FARPROC) _OpenSound = GetProcAddress(driver, MAKEINTRESOURCE(1));
    (FARPROC) _CloseSound = GetProcAddress(driver, MAKEINTRESOURCE(2));
    (FARPROC) _SetVoiceQueueSize = GetProcAddress(driver, MAKEINTRESOURCE(3));
    (FARPROC) _SetVoiceNote = GetProcAddress(driver, MAKEINTRESOURCE(4));
    (FARPROC) _SetVoiceAccent = GetProcAddress(driver, MAKEINTRESOURCE(5));
    (FARPROC) _SetVoiceEnvelope = GetProcAddress(driver, MAKEINTRESOURCE(6));
    (FARPROC) _SetSoundNoise = GetProcAddress(driver, MAKEINTRESOURCE(7));
    (FARPROC) _SetVoiceSound = GetProcAddress(driver, MAKEINTRESOURCE(8));
    (FARPROC) _StartSound = GetProcAddress(driver, MAKEINTRESOURCE(9));
    (FARPROC) _StopSound = GetProcAddress(driver, MAKEINTRESOURCE(10));
    (FARPROC) _WaitSoundState = GetProcAddress(driver, MAKEINTRESOURCE(11));
    (FARPROC) _SyncAllVoices = GetProcAddress(driver, MAKEINTRESOURCE(12));
    (FARPROC) _CountVoiceNotes = GetProcAddress(driver, MAKEINTRESOURCE(13));
    (FARPROC) _GetThresholdEvent = GetProcAddress(driver, MAKEINTRESOURCE(14));
    (FARPROC) _GetThresholdStatus = GetProcAddress(driver, MAKEINTRESOURCE(15));
    (FARPROC) _SetVoiceThreshold = GetProcAddress(driver, MAKEINTRESOURCE(16));
    (FARPROC) _DoBeep = GetProcAddress(driver, MAKEINTRESOURCE(17));
    SetModuleName(driver, "_OLDSND");
    result = 1;
  }
  SetModuleName(_hInstance, "SOUND");
  return result;
}

//*****************************************************************************
// Remove the linkage to the original driver
//*****************************************************************************

void Remove(void)
{
  FreeLibrary(driver);
  (FARPROC) _OpenSound =
  (FARPROC) _CloseSound = 
  (FARPROC) _SetVoiceQueueSize = 
  (FARPROC) _SetVoiceNote = 
  (FARPROC) _SetVoiceAccent = 
  (FARPROC) _SetVoiceEnvelope = 
  (FARPROC) _SetSoundNoise = 
  (FARPROC) _SetVoiceSound = 
  (FARPROC) _StartSound = 
  (FARPROC) _StopSound = 
  (FARPROC) _WaitSoundState = 
  (FARPROC) _SyncAllVoices = 
  (FARPROC) _CountVoiceNotes = 
  (FARPROC) _GetThresholdEvent = 
  (FARPROC) _GetThresholdStatus = 
  (FARPROC) _SetVoiceThreshold = 
  (FARPROC) _DoBeep = 0;
}

//*****************************************************************************
// WEP
//*****************************************************************************

#pragma argsused
int FAR PASCAL WEP(int nParameter)
{
  if (!IsBadReadPtr(&driver, sizeof driver)) Remove();
  return 1;
}

//*****************************************************************************
// LibMain
//*****************************************************************************
// Performs a full initialization only if the original keyboard driver is not
// already loaded on the system.

#pragma argsused
int FAR PASCAL LibMain(HINSTANCE hInstance, WORD wDataSeg, WORD cbHeapSize, LPSTR lpCmdLine)
{
  return IsBadReadPtr(&driver, sizeof(driver)) ? 0 : Install();
}
